@@ -27,7 +27,9 @@ if lib.uninstalled:
     lib.add_global_resource('resource', 'resources')
     lib.add_global_resource('doc', '.')
     lib.add_global_resource('pixmap', 'pixmaps')
+    lib.add_global_resource('pixmap', 'pixmaps/kiwi')
+    lib.add_global_resource('plugins', 'plugins')
 lib.enable_translation()
 
-__version__ = "0.6.6"
+__version__ = "0.7.1"
 
