@@ -17,9 +17,15 @@
 
 import gtk
 
-from gazpacho import util
+from gazpacho import gapi, util
+from gazpacho.command import Command
+from gazpacho.clipboard import CommandCutPaste
 from gazpacho.commandmanager import command_manager
-from gazpacho.widget import Gadget
+from gazpacho.gadget import Gadget
+from gazpacho.i18n import _
+from gazpacho.placeholder import Placeholder
+from gazpacho.widgets.base.box import DragAppendCommand, CreateAppendCommand, \
+     DragExtendCommand, CreateExtendCommand
 
 (DND_POS_TOP,
  DND_POS_BOTTOM,
@@ -63,8 +69,8 @@ class DnDHandler(object):
         widget.drag_source_set(gtk.gdk.BUTTON1_MASK,
                                [DND_WIDGET_TARGET, DND_XML_TARGET],
                                gtk.gdk.ACTION_COPY | gtk.gdk.ACTION_MOVE)
-        widget.connect('drag_begin', self._on_drag_begin)
-        widget.connect('drag_data_get', self._on_drag_data_get)
+        widget.connect('drag_begin', self._on_widget__drag_begin)
+        widget.connect('drag_data_get', self._on_widget__drag_data_get)
 
 
     def connect_drop_handlers(self, widget):
@@ -80,11 +86,11 @@ class DnDHandler(object):
                               DND_XML_TARGET,
                               DND_PALETTE_TARGET],
                              gtk.gdk.ACTION_MOVE | gtk.gdk.ACTION_COPY)
-        widget.connect('drag_motion', self._on_drag_motion)
-        widget.connect('drag_leave', self._on_drag_leave)
-        widget.connect('drag_drop', self._on_drag_drop)
-        widget.connect('drag_data_received', self._on_drag_data_received)
-
+        widget.connect('drag_motion', self._on_widget__drag_motion)
+        widget.connect('drag_leave', self._on_widget__drag_leave)
+        widget.connect('drag_drop', self._on_widget__drag_drop)
+        widget.connect('drag_data_received',
+                       self._on_widget__drag_data_received)
 
 
     #
@@ -145,15 +151,16 @@ class DnDHandler(object):
         # being dragged (for example labels)
         source_widget = drag_context.get_source_widget()
         source_gadget = Gadget.from_widget(source_widget)
-        dragged_widget = source_gadget.dnd_widget.widget
+        dragged_widget = source_gadget.dnd_gadget.widget
 
         # Not a valid drop zone if target == source.
-        if dragged_widget == target_widget:
+        if dragged_widget is target_widget:
             return False
 
         # Make sure we don't try to drop a widget onto one of its
-        # children
-        if dragged_widget.is_ancestor(target_widget):
+        # children or parent
+        if (dragged_widget.is_ancestor(target_widget)
+            or target_widget.is_ancestor(dragged_widget)):
             return False
 
         return True
@@ -180,28 +187,28 @@ class DnDHandler(object):
         necessarily the same widget as the one that recieved the drag
         event.
         """
-        dnd_widget = None
+        dnd_gadget = None
         if info == INFO_TYPE_WIDGET:
             source_widget = drag_context.get_source_widget()
             source_gadget = Gadget.from_widget(source_widget)
-            dnd_widget = source_gadget.dnd_widget
+            dnd_gadget = source_gadget.dnd_gadget
 
         elif info == INFO_TYPE_XML:
-            dnd_widget = Gadget.from_xml(project, data.data)
+            dnd_gadget = Gadget.from_xml(project, data.data)
 
-        return dnd_widget
+        return dnd_gadget
 
 
     #
     # Signal handlers for the drag source
     #
 
-    def _on_drag_begin(self, source_widget, drag_context):
+    def _on_widget__drag_begin(self, source_widget, drag_context):
         """Callback for the 'drag-begin' event."""
         raise NotImplementedError
 
-    def _on_drag_data_get(self, source_widget, drag_context, selection_data,
-                          info, time):
+    def _on_widget__drag_data_get(self, source_widget, drag_context,
+                                  selection_data, info, time):
         """Callback for the 'drag-data-get' event."""
         raise NotImplementedError
 
@@ -210,12 +217,12 @@ class DnDHandler(object):
     # Signal handlers for the drag target
     #
 
-    def _on_drag_leave(self, target_widget, drag_context, time):
+    def _on_widget__drag_leave(self, target_widget, drag_context, time):
         """Callback for the 'drag-leave' event. We clear the drag
         highlight."""
         self._clear_drag_highlight(target_widget)
 
-    def _on_drag_motion(self, target_widget, drag_context, x, y, time):
+    def _on_widget__drag_motion(self, target_widget, drag_context, x, y, time):
         """Callback for the 'drag-motion' event. If the drop zone is
         valid we set the drag highlight and the drag action."""
         if not self._is_valid_drop_zone(drag_context, target_widget):
@@ -231,27 +238,21 @@ class DnDHandler(object):
         drag_context.drag_status(drag_action, time)
         return True
 
-    def _on_drag_drop(self, target_widget, drag_context, x, y, time):
+    def _on_widget__drag_drop(self, target_widget, drag_context, x, y, time):
         """Callback for handling the 'drag_drop' event."""
         if not drag_context.targets:
             return False
 
-        if drag_context.get_source_widget():
-            # For DnD within the application we request to use the
-            # widget or adaptor directly
-            if MIME_TYPE_OBJECT_PALETTE in drag_context.targets:
-                mime_type = MIME_TYPE_OBJECT_PALETTE
-            else:
-                mime_type = MIME_TYPE_OBJECT
+        if MIME_TYPE_OBJECT_PALETTE in drag_context.targets:
+            mime_type = MIME_TYPE_OBJECT_PALETTE
+        elif drag_context.action == gtk.gdk.ACTION_MOVE:
+            mime_type = MIME_TYPE_OBJECT
         else:
-            # otherwise we'll have request the data to be passed as XML
             mime_type = MIME_TYPE_OBJECT_XML
-
         target_widget.drag_get_data(drag_context, mime_type, time)
-
         return True
 
-    def _on_drag_data_received(self, target_widget, drag_context,
+    def _on_widget__drag_data_received(self, target_widget, drag_context,
                                   x, y, data, info, time):
         """Callback for the 'drag-data-recieved' event."""
         raise NotImplementedError
@@ -397,35 +398,36 @@ class WidgetDnDHandler(DnDHandler):
             pos = pos + 1
         return pos
 
-    def _execute_drag(self, location, dnd_widget, target_gadget, keep_source):
+    def _execute_drag(self, location, dnd_gadget, target_gadget, keep_source):
         """
         Execute the drag-append or drag-extend command.
 
         @param location: the drop location
         @type location: dnd constant
-        @param dnd_widget: the widget that is being dragged
-        @type dnd_widget: L{gazpacho.widget.Widget}
-        @param target_gadget: the widget where the drop occurred
-        @type target_gadget: L{gazpacho.widget.Widget}
+        @param dnd_gadget: the gadget that is being dragged
+        @type dnd_gadget: L{gazpacho.gadget.Gadget}
+        @param target_gadget: the gadget where the drop occurred
+        @type target_gadget: L{gazpacho.gadget.Gadget}
         @param keep_source: True if the source widget should not be removed
         @type keep_source: bool
 
         @return: True if the drop was successful
         @rtype: bool
         """
-        if not dnd_widget:
+        if not dnd_gadget:
             return False
 
         if self._is_extend_action(location):
-            command_manager.execute_drag_extend(dnd_widget, target_gadget,
-                                                location, keep_source)
+            cmd = DragExtendCommand(dnd_gadget, target_gadget, location,
+                                    keep_source)
+            command_manager.execute(cmd, target_gadget.project)
             return True
 
         elif self._is_append_action(location):
             pos = self._get_append_position(target_gadget.widget, location)
             parent = target_gadget.get_parent()
-            command_manager.execute_drag_append(dnd_widget, parent, pos,
-                                                keep_source)
+            cmd = DragAppendCommand(dnd_gadget, parent, pos, keep_source)
+            command_manager.execute(cmd, dnd_gadget.project)
             return True
 
         return False
@@ -438,75 +440,85 @@ class WidgetDnDHandler(DnDHandler):
         @type location: dnd constant
         @param adaptor: adaptor for creating the class
         @type adaptor: L{gazpacho.widgetadaptor.WidgetAdaptor}
-        @param target_gadget: the widget where the drop occurred
-        @type target_gadget: L{gazpacho.widget.Widget}
+        @param target_gadget: the gadget where the drop occurred
+        @type target_gadget: L{gazpacho.gadget.Gadget}
 
         @return: True if the drop was successful
         @rtype: bool
         """
+        project = target_gadget.project
+        success = False
+
         if self._is_extend_action(location):
-            command_manager.execute_create_extend(adaptor, target_gadget,
-                                                  location)
-            return True
+            gadget = Gadget(adaptor, project)
+            gadget.create_widget(interactive=True)
+            cmd = CreateExtendCommand(gadget, target_gadget, location)
+            command_manager.execute(cmd, project)
+
+            success = True
 
         elif self._is_append_action(location):
             pos = self._get_append_position(target_gadget.widget, location)
             parent = target_gadget.get_parent()
-            command_manager.execute_create_append(adaptor, parent, pos)
-            return True
+            gadget = Gadget(adaptor, project)
+            gadget.create_widget(interactive=True)
+            cmd = CreateAppendCommand(gadget, parent, pos)
+            command_manager.execute(cmd, project)
 
-        return False
+            success = True
 
+        if success:
+            from gazpacho.palette import palette
+            if not palette.persistent_mode:
+                palette.unselect_widget()
+
+        return success
 
     #
     # Signal handlers for the drag source
     #
 
-    def _on_drag_begin(self, source_widget, drag_context):
+    def _on_widget__drag_begin(self, widget, drag_context):
         """
         Set a drag icon that matches the drag source widget.
         """
-        source_gadget = Gadget.from_widget(source_widget)
-        if source_gadget.dnd_widget:
-            pixbuf = source_gadget.dnd_widget.adaptor.icon.get_pixbuf()
-            source_widget.drag_source_set_icon_pixbuf(pixbuf)
+        gadget = Gadget.from_widget(widget)
+        if gadget.dnd_gadget:
+            pixbuf = gadget.dnd_gadget.adaptor.icon.get_pixbuf()
+            widget.drag_source_set_icon_pixbuf(pixbuf)
 
-    def _on_drag_data_get(self, source_widget, drag_context, selection_data,
-                          info, time):
+    def _on_widget__drag_data_get(self, widget, drag_context,
+                                  selection_data, info, time):
         """
         Make the widget data available in the format that was
         requested. If the drag and drop occurs within the application
         the widget can be accessed directly otherwise it has to be
         passed as an XML string.
         """
-        source_gadget = Gadget.from_widget(source_widget)
-        dnd_widget = source_gadget.dnd_widget
+
+        gadget = Gadget.from_widget(widget)
 
         # If we can't get the widget we indicate this failure by
         # passing an empty string. Not sure if it's correct but it
-        # works for us.
-        if not dnd_widget:
-            selection_data.set(selection_data.target, 8, "")
-            return
-
-        # The widget should be passed as XML
-        if info == INFO_TYPE_XML:
-            selection_data.set(selection_data.target, 8, dnd_widget.to_xml())
-
-        # The widget can be retrieved directly and we only pass the name
-        elif info == INFO_TYPE_WIDGET:
-            selection_data.set(selection_data.target, 8, dnd_widget.name)
-        # We don't understand the request and pass nothing
-        else:
-            selection_data.set(selection_data.target, 8, "")
+        # works for us. Note that the source widget is sometimes
+        # different from the dragged widget.
+        data = ""
+        if gadget.dnd_gadget:
+            # The widget should be passed as XML
+            if info == INFO_TYPE_XML:
+                data = gadget.dnd_gadget.to_xml(skip_external_references=True)
+            # The widget can be retrieved directly and we only pass the name
+            elif info == INFO_TYPE_WIDGET:
+                data = gadget.dnd_gadget.name
 
+        selection_data.set(selection_data.target, 8, data)
 
     #
     # Signal handlers for the drag target
     #
 
-    def _on_drag_data_received(self, target_widget, drag_context, x, y, data,
-                               info, time):
+    def _on_widget__drag_data_received(self, target_widget, drag_context,
+                                       x, y, data, info, time):
         """
         The data has been received and the appropriate command can now
         be executed.
@@ -516,42 +528,43 @@ class WidgetDnDHandler(DnDHandler):
         handles the drop (paste) and the source takes care of the drag
         (cut).
         """
-        # If there is no data we indicate that the drop was not
-        # successful
-        if not data.data:
-            drag_context.finish(False, False, time)
-            return
-
-        project = self._get_target_project(target_widget)
-        location = self._get_drop_location(target_widget, x, y)[0]
-        target_gadget = Gadget.from_widget(target_widget)
 
+        # If there is no data we indicate that the drop was not successful
         success = False
-        if info in (INFO_TYPE_WIDGET, INFO_TYPE_XML):
-            dnd_gadget = self._get_dnd_gadget(data, info, drag_context, project)
-            keep_source = (drag_context.action != gtk.gdk.ACTION_MOVE)
-            success = self._execute_drag(location, dnd_gadget, target_gadget,
-                                         keep_source)
+        if data.data:
+            project = self._get_target_project(target_widget)
+            location = self._get_drop_location(target_widget, x, y)[0]
+            target_gadget = Gadget.from_widget(target_widget)
 
-        elif info == INFO_TYPE_PALETTE:
-            adaptor = project.get_app().add_class
-            success = self._execute_create(location, adaptor, target_gadget)
+            if info in (INFO_TYPE_WIDGET, INFO_TYPE_XML):
+                dnd_gadget = self._get_dnd_gadget(data, info,
+                                                  drag_context, project)
+                keep_source = (drag_context.action != gtk.gdk.ACTION_MOVE)
+                success = self._execute_drag(location, dnd_gadget,
+                                             target_gadget, keep_source)
+
+            elif info == INFO_TYPE_PALETTE:
+                adaptor = project.get_app().add_class
+                if adaptor:
+                    success = self._execute_create(location, adaptor,
+                                                   target_gadget)
+                else:
+                    success = False
 
         drag_context.finish(success, False, time)
 
-
 class PlaceholderDnDHandler(DnDHandler):
 
-    def _get_target_project(self, target_widget):
+    def _get_target_project(self, widget):
         """
         Get the project to which the target belongs.
 
-        @param target_widget: the target placeholder
-        @type target_widget: L{gazpacho.placeholder.Placeholder}
+        @param widget: the target placeholder
+        @type widget: L{gazpacho.placeholder.Placeholder}
         """
-        return target_widget.get_project()
+        return widget.get_project()
 
-    def _set_drag_highlight(self, target_widget, x, y):
+    def _set_drag_highlight(self, widget, x, y):
         """
         Highlight the widget in an appropriate way to indicate
         that this is a valid drop zone.
@@ -563,24 +576,24 @@ class PlaceholderDnDHandler(DnDHandler):
         @param y: the mouse y position
         @type y: int
         """
-        target_widget.drag_highlight()
+        widget.drag_highlight()
 
-    def _clear_drag_highlight(self, target_widget):
+    def _clear_drag_highlight(self, widget):
         """
         Clear the drag highligt.
 
-        @param target_widget: the gtk target widget
-        @type target_widget: gtk.Widget
+        @param widget: the target widget
+        @type widget: gtk.Widget
         """
-        target_widget.drag_unhighlight()
+        widget.drag_unhighlight()
 
-    def _execute_drag(self, dnd_gadget, placeholder, keep_source, project):
+    def _execute_drag(self, gadget, placeholder, keep_source, project):
         """
         Execute the drag and drop command.
 
-        @param dnd_gadget: the gadget that is being dragged
-        @type dnd_gadget: L{gazpacho.widget.Gadget}
-        @param placeholder: the widget where the drop occurred
+        @param gadget: the gadget that is being dragged
+        @type gadget: L{gazpacho.gadget.Gadget}
+        @param placeholder: the placeholder where the drop occurred
         @type placeholder: L{gazpacho.placeholder.Placeholder}
         @param keep_source: True if the source widget should not be removed
         @type keep_source: bool
@@ -590,23 +603,24 @@ class PlaceholderDnDHandler(DnDHandler):
         @return: True if the drop was successful
         @rtype: bool
         """
-        if not dnd_gadget:
+        if not gadget:
             return False
 
         if keep_source:
-            command_manager.execute_drop(dnd_gadget.copy(project),
-                                         placeholder)
+            cmd = CommandCutPaste(gadget.copy(project), project, placeholder,
+                                  False)
         else:
-            command_manager.execute_drag_drop(dnd_gadget, placeholder)
+            cmd = CommandDragDrop(gadget, placeholder)
 
+        command_manager.execute(cmd, project)
         return True
 
     #
     # Signal handlers for the drag target
     #
 
-    def _on_drag_data_received(self, target_widget, drag_context, x, y, data,
-                               info, time):
+    def _on_widget__drag_data_received(self, target_widget, drag_context,
+                                       x, y, data, info, time):
         """
         The data has been received and the appropriate command can now
         be executed.
@@ -634,8 +648,80 @@ class PlaceholderDnDHandler(DnDHandler):
             success = self._execute_drag(dragged_gadget, target_widget,
                                          keep_source, project)
         elif info == INFO_TYPE_PALETTE:
-            command_manager.create(project.get_app().add_class, target_widget,
-                                   None)
+            gapi.create_gadget(project, project.get_app().add_class,
+                               target_widget)
             success = True
 
         drag_context.finish(success, False, time)
+
+
+class CommandDragDrop(Command):
+    """Command for executing a drag and drop action. This will move
+    the widget and cannot be used to drag widgets between programs.
+    """
+
+    def __init__(self, source_gadget, target_placeholder):
+        """
+        Initialize the command.
+
+        @param source_gadget: the source gadget
+        @type source_gadget: gazpacho.gadget.Gadget
+        @param target_placeholder: the target placeholder
+        @type target_placeholder: gazpacho.placeholder.Placeholder
+        """
+        description = _("Drag and Drop widget %s") % source_gadget.name
+        Command.__init__(self, description)
+
+        self._undo = False
+        self._source_placeholder = None
+        self._target_placeholder = target_placeholder
+        self._source = source_gadget
+
+    def execute(self):
+        if self._undo:
+            self._remove_gadget(self._source, self._target_placeholder)
+            self._add_gadget(self._source, self._source_placeholder)
+
+        else:
+            # Delay the construction of the placeholder until we need it
+            if not self._source_placeholder:
+                self._source_placeholder = Placeholder()
+
+            self._remove_gadget(self._source, self._source_placeholder)
+            self._add_gadget(self._source, self._target_placeholder)
+
+        self._undo = not self._undo
+
+    def _remove_gadget(self, gadget, placeholder):
+        """
+        Remove the widget from the gadget and replace it with a
+        new placeholder.
+
+        @param gadget: the gadget that should be removed
+        @type gadget: gazpacho.gadget.Gadget
+        @param placeholder: the new placeholder
+        @type placeholder: gazpacho.placeholder.Placeholder
+        """
+        parent = gadget.get_parent()
+        Gadget.replace(gadget.widget, placeholder, parent)
+
+        gadget.widget.hide()
+        gadget.project.remove_widget(gadget.widget)
+
+    def _add_gadget(self, gadget, placeholder):
+        """
+        Remove the old placeholder and add a widget from the gadget.
+
+        @param gadget: the gadget that should be added
+        @type gadget: gazpacho.gadget.Gadget
+        @param placeholder: the old placeholder
+        @type placeholder: gazpacho.placeholder.Placeholder
+        """
+        parent = util.get_parent(placeholder)
+        project = parent.project
+        Gadget.replace(placeholder, gadget.widget, parent)
+
+        project.add_widget(gadget.widget)
+        gadget.select()
+
+        gadget.widget.show_all()
