@@ -14,139 +14,91 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 
-import gettext
 import os.path
 
 import gtk
 import gobject
+from kiwi.utils import gsignal, type_register
 
-from gazpacho import util
+from gazpacho import gapi, util
 from gazpacho.commandmanager import command_manager
 from gazpacho.context import Context
-from gazpacho.kiwiutils import gsignal, type_register
-from gazpacho.loader.loader import ObjectBuilder
-from gazpacho.loader.custom import ButtonAdapter, DialogAdapter, \
-     WindowAdapter, str2bool, adapter_registry
+from gazpacho.filewriter import XMLWriter
+from gazpacho.gadget import Gadget, load_gadget_from_widget
+from gazpacho.loader.gazpacholoader import GazpachoObjectBuilder
+from gazpacho.model import GazpachoModelManager
 from gazpacho.placeholder import Placeholder
 from gazpacho.signaleditor import SignalInfo
 from gazpacho.sizegroup import GSizeGroup
-from gazpacho.uim import GazpachoUIM
-from gazpacho.widget import Gadget, load_widget_from_widget
+from gazpacho.uimanager import GazpachoUIM
+from gazpacho.widgets.base.box import CommandBoxDeletePlaceholder
 from gazpacho.widgetregistry import widget_registry
-from gazpacho.widgets.base.custom import Custom
-from gazpacho.filewriter import XMLWriter
-
-_ = gettext.gettext
-
-class GazpachoObjectBuilder(ObjectBuilder):
-    def __init__(self, **kwargs):
-        self._app = kwargs.pop('app', None)
-        kwargs['placeholder'] = self.create_placeholder
-        kwargs['custom'] = Custom
-        ObjectBuilder.__init__(self, **kwargs)
-
-    def create_placeholder(self, name):
-        return Placeholder()
-
-    def show_windows(self):
-        pass
-
-class GazpachoButtonAdapter(ButtonAdapter):
-    def add(self, parent, child, properties):
-        # This is a hack, remove when the button saving/loading code
-        # is improved
-        for cur_child in parent.get_children():
-            parent.remove(cur_child)
-        ButtonAdapter.add(self, parent, child, properties)
-
-
-class GazpachoWindowAdapter(WindowAdapter):
-    object_type = gtk.Window
-    def prop_set_visible(self, window, value):
-        window.set_data('gazpacho::visible', str2bool(value))
-
-    def prop_set_modal(self, window, value):
-        window.set_data('gazpacho::modal', str2bool(value))
-
-adapter_registry.register_adapter(GazpachoWindowAdapter)
-
-# FIXME: This is quite strange, but apparently we need this too.
-class GazpachoDialogAdapter(DialogAdapter):
-    object_type = gtk.Dialog
-    def prop_set_modal(self, window, value):
-        window.set_data('gazpacho::modal', str2bool(value))
-
-adapter_registry.register_adapter(GazpachoDialogAdapter)
 
 class Project(gobject.GObject):
-    gsignal('add-widget', object),
-    gsignal('remove-widget', object),
-    gsignal('project-changed'),
-    gsignal('selection-changed'),
+    """
+    Project is an object representing a specific user interface which
+    can be saved on disk.
+
+    Signals:
+      changed: something in the project changes, used to set the save button
+        insensitive for example
+      add-gadget: a gadget was added
+      remove-gadget: a gadget was removed
+      gadget-name-changed: the name of a gadget changed, different from
+        when the name in a widget changes, because in some cases the widget
+        itself changes in a gadget
+      add-action: an action or actiongroup was added
+      remove-action: an action or actiongroup was removed
+      action-name-changed: the name of an action changed
+      add-sizegroup: a sizegroup was added
+      remove-sizegroup: a sizegroup was removed
+
+    @ivar name: The name of the project like network-conf
+    @ivar path: The full path of the xml file for this project
+    @ivar sizegroups: # A list of GSizeGroups that are part of this project
+    @ivar selection: We need to keep the selection in the project because
+      we have multiple projects and when the user switchs between them, he
+      will probably not want to loose the selection. This is a list of
+      widget items.
+    @ivar uim: There is a UIManager in each project which holds the information
+      about menus and toolbars
+    @ivar model_manager: The model manager takes care of all the data models
+      for TreeViews and ComboBoxes, ...
+    @ivar undo_stack: Contains the undo (and redo) commands
+    @ivar context: out context
+    """
+    gsignal('changed'),
+    gsignal('add-gadget', object),
+    gsignal('remove-gadget', object),
+    gsignal('gadget-name-changed', object),
     gsignal('add-action', object),
     gsignal('remove-action', object),
+    gsignal('action-name-changed', object),
     gsignal('add-sizegroup', object),
     gsignal('remove-sizegroup', object),
 
-    # XXX: notify::name
-    gsignal('widget-name-changed', object),
-    gsignal('action_name_changed', object),
-
-    project_counter = 1
-
-    def __init__(self, untitled, app, loader=None):
+    def __init__(self, app):
         gobject.GObject.__init__(self)
 
         self._app = app
-
-        self.name = None
-
-        if untitled:
-            # The name of the project like network-conf
-            self.name = _('Untitled %d') % Project.project_counter
-            Project.project_counter += 1
-
-        # The full path of the xml file for this project
-        self.path = None
-
-        # The menu entry in the /Project menu
-        self.entry = None
-
+        self._changed = False
+        self._domain = ''
+        self._unsupported_widgets = {}
+        self._version = 'libglade'
+        self._widget_old_names = {} # widget -> old name of the
         self._widgets = {}
 
-        # A list of GSizeGroups that are part of this project
+        self.name = None
+        self.path = None
         self.sizegroups = []
-
-        # We need to keep the selection in the project because we have multiple
-        # projects and when the user switchs between them, he will probably
-        # not want to loose the selection. This is a list of GtkWidget items.
-        self.selection = []
-
+        self.selection = WidgetSelection()
         self.tooltips = gtk.Tooltips()
-
-        # A flag that is set when a project has changes if this flag is not set
-        # we don't have to query for confirmation after a close or exit is
-        # requested
-        self._changed = False
-
-        # There is a UIManager in each project which holds the information
-        # about menus and toolbars
         self.uim = GazpachoUIM(self)
-
-        # This is the id for the menuitem entry on Gazpacho main menu for this
-        # project
-        self.uim_id = -1
-
-        # Contains the undo (and redo) commands
+        self.model_manager = GazpachoModelManager()
         self.undo_stack = UndoRedoStack()
         self.undo_stack.connect('changed', self._undo_stack_changed_cb)
-
-        # create our context
         self.context = Context(self)
 
-        if loader:
-            self._read_from_loader(loader)
-
     # Private
 
     def _read_from_loader(self, loader):
@@ -157,12 +109,19 @@ class Project(gobject.GObject):
         @param loader: the loader
         @type loader: L{gazpacho.loader.loader.ObjectBuilder} instance
         """
+        self._domain = loader.get_domain() or ''
+        self._version = loader.get_version()
 
         # Load UIM manager
         # Do this first since the custom menubar and toolbars adapters
         # depends on all ui definitions being loaded
         self.uim.load(loader)
 
+        # Load models before widgets
+        models = [w for w in loader.toplevels if isinstance(w, gtk.ListStore)]
+        for model in models:
+            self.model_manager.load_model(model)
+
         # Load the widgets
         for widget in loader.toplevels:
             if isinstance(widget, gtk.Widget):
@@ -172,7 +131,7 @@ class Project(gobject.GObject):
         # since the sizegroups has references to the widgets
         for sizegroup in loader.sizegroups:
             name = sizegroup.get_data('gazpacho::object-id')
-            widgets =  sizegroup.get_data('gazpacho::sizegroup-widgets')
+            widgets =  sizegroup.get_data('gazpacho::sizegroup-widgets') or []
             gadgets = [Gadget.from_widget(widget)
                        for widget in widgets]
             self.add_sizegroup(GSizeGroup(name, sizegroup, gadgets))
@@ -187,10 +146,12 @@ class Project(gobject.GObject):
                                                  handler=signal_handler,
                                                  after=signal_after))
 
+        self._unsupported_widgets = loader.get_unsupported_widgets()
+
         self.changed = False
 
     def _load_widget(self, widget):
-        gadget = load_widget_from_widget(widget, self)
+        gadget = load_gadget_from_widget(widget, self)
         self.add_widget(widget)
         return gadget
 
@@ -204,24 +165,29 @@ class Project(gobject.GObject):
     def _remove_action_cb(self, action_group, action):
         self.emit('remove-action', action)
 
-#     def _on_gadget_notify_name(self, gadget, pspec):
-#         self.emit('widget-name-changed', gadget)
+    def _on_widget_notify_name(self, widget, pspec):
+        old_name = self._widget_old_names[widget]
+        new_name = widget.get_name()
+        if old_name == new_name:
+            # This happens but I'm not sure why, it would be nice if
+            # it could be avoided
+            return
 
-#         if isinstance(gadget.widget, (gtk.MenuBar, gtk.Toolbar)):
-#             self.uim.update_widget_name(gadget)
-#             self.replace(old_name, gadget.widget)
+        del self._widget_old_names[widget]
+        del self._widgets[old_name]
 
-    # Class methods
+        gadget = Gadget.from_widget(widget)
+        if isinstance(widget, (gtk.MenuBar, gtk.Toolbar)):
+            self.uim.update_widget_name(gadget)
+            # gadget.widget might have been replaced with a new widget
+            # after calling this method
+            widget = gadget.widget
+            widget.connect('notify::name', self._on_widget_notify_name)
 
-    def open(cls, path, app, buffer=None):
-        loader = GazpachoObjectBuilder(filename=path, buffer=buffer, app=app)
+        self._widget_old_names[widget] = new_name
+        self._widgets[widget.get_name()] = widget
 
-        project = cls(False, app, loader)
-        project.path = path
-        project.name = os.path.basename(path)
-
-        return project
-    open = classmethod(open)
+        self.emit('gadget-name-changed', gadget)
 
     # Properties
 
@@ -231,11 +197,72 @@ class Project(gobject.GObject):
     def _set_changed(self, value):
         if self._changed != value:
             self._changed = value
-            self.emit('project-changed')
+            self.emit('changed')
     changed = property(_get_changed, _set_changed)
 
     # Public API
 
+    def load(self, filename):
+        """
+        Loads a glade file from a filename and inserts the content of
+        the file in a project.
+
+        @param filename: glade file to open
+        """
+        loader = GazpachoObjectBuilder(filename=filename, app=self._app)
+        self._read_from_loader(loader)
+        self.path = filename
+        self.name = os.path.basename(filename)
+
+    def load_from_buffer(self, buffer):
+        """
+        Loads a glade buffer and inserts the content of the file in a project.
+
+        @param filename: string
+        """
+        loader = GazpachoObjectBuilder(buffer=buffer, app=self._app)
+        self._read_from_loader(loader)
+
+    def get_domain(self):
+        """
+        Retrieves the translatable domain
+
+        @returns: the domain
+        """
+        return self._domain
+
+    def set_domain(self, domain):
+        """
+        Set the translatable domain for this glade file
+
+        @param domain: domain to set
+        @type domain: string
+        """
+
+        self._domain = domain
+
+        self.changed = True
+
+    def get_version(self):
+        """
+        Retrieves the version, possible values are currently
+        'libglade' or 'gtkbuilder'
+
+        @returns: the version
+        """
+        return self._version
+
+    def get_unsupported_widgets(self):
+        """
+        Get a dict of the widgets that could not be loaded because
+        they're not supported. The dict maps the class of the widget
+        to a list of widget ids.
+
+        @return: the unsupported widgets
+        @rtype: dict
+        """
+        return self._unsupported_widgets
+
     def get_app(self):
         return self._app
 
@@ -258,47 +285,43 @@ class Project(gobject.GObject):
     def add_widget(self, widget, new_name=False):
         """
         Adds a widget to the project and optionally sets a new name
-        if it already exists. Will emit add-widget signal when done.
+        if it already exists. Will emit add-gadget signal when done.
 
         @param widget:
         @param new_name: If True assign a new name if it already
           exists
         """
-
         # we don't list placeholders
         if isinstance(widget, Placeholder):
             return
 
-        if isinstance(widget, gtk.Container):
-            children = widget.get_children()
-        else:
-            children = []
-
         # The internal widgets (e.g. the label of a GtkButton) are handled
-        # by gtk and don't have an associated GladeGadget: we don't want to
+        # by gtk and don't have an associated gadget: we don't want to
         # add these to our list. It would be nicer to have a flag to check
         # (as we do for placeholders) instead of checking for the associated
-        # GladeWidget, so that we can assert that if a widget is _not_
-        # internal, it _must_ have a corresponding GladeWidget...
+        # gadget, so that we can assert that if a widget is _not_
+        # internal, it _must_ have a corresponding gadget...
         # Anyway this suffice for now.
         gadget = Gadget.from_widget(widget)
         if not gadget:
             return
 
+        children = gadget.get_children()
+
         gadget.project = self
+        self._widget_old_names[widget] = widget.get_name()
 
         widget_name = widget.name
         if new_name and gadget.name in self._widgets:
             widget_name = self.set_new_widget_name(gadget.widget)
 
-        # XXX: Currently broken
-        #widget.connect('notify::name', self._on_gadget_notify_name)
+        widget.connect('notify::name', self._on_widget_notify_name)
         self._widgets[widget_name] = widget
 
         for child in children:
             self.add_widget(child, new_name=new_name)
 
-        self.emit('add-widget', gadget)
+        self.emit('add-gadget', gadget)
 
     def add_hidden_widget(self, widget):
         """
@@ -306,31 +329,58 @@ class Project(gobject.GObject):
         It also updates the name of the widget
         @param widget: the widget
         """
-
         if not isinstance(widget, gtk.Widget):
             raise TypeError("widget %r is not gtk.Widget instance" % widget)
 
+        if isinstance(widget, gtk.Container):
+            for child in util.get_all_children(widget):
+                self.add_hidden_widget(child)
+
         self._widgets[widget.get_name()] = widget
 
+    def remove_hidden_widget(self, widget):
+        if not isinstance(widget, gtk.Widget):
+            raise TypeError("widget %r is not gtk.Widget instance" % widget)
+
+        if isinstance(widget, gtk.Container):
+            for child in util.get_all_children(widget):
+                self.remove_hidden_widget(child)
+
+        if widget.get_name() in self._widgets.keys():
+            del self._widgets[widget.get_name()]
+
     def remove_widget(self, widget):
         if isinstance(widget, Placeholder):
             return
 
-        if isinstance(widget, gtk.Container):
-            for child in widget.get_children():
-                self.remove_widget(child)
-
         gadget = Gadget.from_widget(widget)
         if not gadget:
             return
 
+        for child in gadget.get_children():
+            self.remove_widget(child)
+
         if widget in self.selection:
-            self.selection_remove(widget, False)
-            self.selection_changed()
+            self.selection.remove(widget, False)
+            self.selection.selection_changed()
 
         del self._widgets[widget.get_name()]
 
-        self.emit('remove-widget', gadget)
+        self.emit('remove-gadget', gadget)
+
+    def delete_selection(self):
+        """Delete the widget in the selection calling the right command"""
+        assert len(self.selection) == 1
+
+        widget = self.selection[0]
+        gadget = Gadget.from_widget(widget)
+        if gadget:
+            gapi.delete_gadget(self, gadget)
+
+        elif (isinstance(widget, Placeholder) and widget.is_deletable()):
+            if len(widget.get_parent().get_children()) > 1:
+                cmd = CommandBoxDeletePlaceholder(widget)
+                command_manager.execute(cmd, self)
 
     # XXX: replace_widget
     def replace(self, widget_name, new_widget):
@@ -387,72 +437,10 @@ class Project(gobject.GObject):
     def change_action_name(self, action):
         self.emit('action-name-changed', action)
 
-    def selection_changed(self):
-        self.emit('selection-changed')
-
-    def selection_clear(self, emit_signal):
-        if not self.selection:
-            return
-
-        for widget in self.selection:
-            util.remove_nodes(widget)
-
-        self.selection = []
-
-        if emit_signal:
-            self.selection_changed()
-
-    # XXX: remove_selected
-    def selection_remove(self, widget, emit_signal):
-        """ Remove the widget from the current selection """
-        if not util.has_nodes(widget):
-            return
-
-        util.remove_nodes(widget)
-
-        self.selection.remove(widget)
-        if emit_signal:
-            self.selection_changed()
-
-    # XXX: add_selected
-    def selection_add(self, widget, emit_signal):
-        if util.has_nodes(widget):
-            return
-
-        util.add_nodes(widget)
-
-        self.selection.insert(0, widget)
-        if emit_signal:
-            self.selection_changed()
-
-    # XXX: select_widget
-    def selection_set(self, widget, emit_signal):
-        # If the widget is already selected we don't have to bother
-        if util.has_nodes(widget):
-            # But we need to emit the selection_changed signal if the
-            # widget isn't in the active project
-            if self._app._project != self and emit_signal:
-                self.selection_changed()
-            return
-
-        self.selection_clear(False)
-        self.selection_add(widget, emit_signal)
-
-    def delete_selection(self):
-        """ Delete (remove from the project) the widgets in the selection """
-        newlist = list(self.selection)
-        for widget in newlist:
-            gadget = Gadget.from_widget(widget)
-            if gadget is not None:
-                command_manager.delete(gadget)
-            elif (isinstance(widget, Placeholder)
-                  and widget.is_deletable()):
-                command_manager.delete_placeholder(widget)
-
     def set_new_widget_name(self, widget):
         """Create a new name based on type."""
-        if not isinstance(widget, gtk.Widget):
-            raise TypeError("widget must be a gtk.Widget, not %r" % widget)
+        if not isinstance(widget, gobject.GObject):
+            raise TypeError("Instance must be a GObject, not %r" % widget)
 
         adapter = widget_registry.get_by_type(widget)
         base_name = adapter.generic_name
@@ -467,23 +455,27 @@ class Project(gobject.GObject):
         return name
 
     def save(self, path):
-        self.path = path
-        self.name = os.path.basename(path)
-
         xw = XMLWriter(project=self)
         widgets = self._widgets.values()
         widgets.sort(lambda x, y: cmp(x.name, y.name))
-        retval = xw.write(path, widgets, self.sizegroups, self.uim)
+        models = self.model_manager.get_models()
+        retval = xw.write(path, widgets, self.sizegroups, models, self.uim,
+                          self._domain, self._version)
+
+        self.undo_stack.mark_saved()
+        self.path = path
+        self.name = os.path.basename(path)
 
         self.changed = False
-        self.undo_stack.mark_saved()
+
         return retval
 
     def serialize(self):
         xw = XMLWriter()
         widgets = self._widgets.values()
         widgets.sort(lambda x, y: cmp(x.name, y.name))
-        return xw.serialize_widgets(widgets, self.sizegroups, self.uim)
+        models = self.model_manager.get_models()
+        return xw.serialize_widgets(widgets, self.sizegroups, models, self.uim)
 
 type_register(Project)
 
@@ -589,3 +581,181 @@ class UndoRedoStack(gobject.GObject):
         return self._commands[self._current_undo_command + 1:]
 
 type_register(UndoRedoStack)
+
+def refresh_selected_nodes(widget):
+    """ Invalidates the portion of the window covered by the selection
+    nodes for the given widget """
+
+    if not widget.window:
+        return
+
+    if widget.get_parent():
+        window = widget.get_parent_window()
+    else:
+        window = widget.window
+    window.invalidate_rect(widget.allocation, True)
+
+class WidgetSelection(gobject.GObject):
+    """
+    The widget selection keeps track of which widgets are currently
+    selected and provide means to manipulate the selection.
+
+    When the selection is modified a 'selection-changed' signal will
+    be emitted.
+    """
+
+    gsignal('selection-changed')
+
+    def __init__(self):
+        gobject.GObject.__init__(self)
+        self._selection = []
+
+    def selection_changed(self):
+        """
+        Emit a selection-changed signal.
+        """
+        self.emit('selection_changed')
+
+    def clear(self, emit_signal=True):
+        """
+        Clear the selection. This might emit a selection-changed
+        signal.
+
+        @param emit_signal: if the signal should be emitted
+        @type emit_signal: bool
+        """
+        if not self._selection:
+            return
+
+        for widget in self._selection:
+            refresh_selected_nodes(widget)
+
+        self._selection = []
+
+        if emit_signal:
+            self.selection_changed()
+
+    def remove(self, widget, emit_signal=True):
+        """
+        Remove the widget from the current selection. This might emit
+        a selection-changed signal.
+
+        @param widget: the widget to remove
+        @type widget: gtk.Widget
+        @param emit_signal: if the signal should be emitted
+        @type emit_signal: bool
+        """
+        if widget not in self._selection:
+            return
+
+        refresh_selected_nodes(widget)
+
+        self._selection.remove(widget)
+        if emit_signal:
+            self.selection_changed()
+
+    def add(self, widget, emit_signal=True):
+        """
+        Add the widget to the selection. This might emit a
+        selection-changed signal.
+
+        @param widget: the widget to add
+        @type widget: gtk.Widget
+        @param emit_signal: if the signal should be emitted
+        @type emit_signal: bool
+        """
+        if widget in self._selection:
+            return
+
+        self._selection.insert(0, widget)
+        if emit_signal:
+            self.selection_changed()
+
+        refresh_selected_nodes(widget)
+
+    def set(self, widget, emit_signal=True):
+        """
+        Clear the selection and add the specified widget. This might
+        emit a selection-changed signal.
+
+        @param widget: the widget to add
+        @type widget: gtk.Widget
+        @param emit_signal: if the signal should be emitted
+        @type emit_signal: bool
+        """
+        self.clear(False)
+        self.add(widget, emit_signal)
+
+    def toggle(self, widget, emit_signal=True):
+        """
+        The the widget is in the selection it will be removed
+        otherwise it will be added.
+
+        @param widget: the widget to add or remove
+        @type widget: gtk.Widget
+        @param emit_signal: if the signal should be emitted
+        @type emit_signal: bool
+        """
+        if widget in self._selection:
+            self.remove(widget, emit_signal)
+        else:
+            self.add(widget, emit_signal)
+
+    def circle(self, leaf_widget):
+        """
+        Select the parent of the selected widget. If the selected
+        widget is a toplevel widget or there isn't a selected widget
+        in the current branch of the tree, select the leaf widget.
+
+        @param leaf_widget: the leaf widget
+        @type leaf_widget: gtk.Widget
+        """
+        parent_gadget = util.get_parent(leaf_widget)
+
+        # If the leaf widget doesn't have a parent select it.
+        if not parent_gadget:
+            self.set(leaf_widget)
+            return
+
+        # If there is no or multiple selections, select the leaf
+        # widget. Note that it's not sure that the leaf widget has a
+        # gadget (if it' a placeholder).
+        if len(self._selection) != 1:
+            self.set(leaf_widget)
+            return
+
+        selected = self._selection[0]
+
+        # If the leaf widget is selected, select its parent
+        if leaf_widget is selected:
+            self.set(parent_gadget.widget)
+            return
+
+        # Circle through the tree
+        new_selection = leaf_widget
+        while parent_gadget:
+            if parent_gadget.widget is selected:
+                # Choose the leaf if selected is toplevel
+                if parent_gadget.is_toplevel():
+                    new_selection = leaf_widget
+                # otherwise choose the parent
+                else:
+                    new_selection = parent_gadget.get_parent().widget
+                break
+            parent_gadget = parent_gadget.get_parent()
+        self.set(new_selection)
+
+
+    def __iter__(self):
+        return iter(self._selection)
+
+    def __len__(self):
+        return len(self._selection)
+
+    def __contains__(self, item):
+        return item in self._selection
+
+    def __getitem__(self, key):
+        return self._selection[key]
+
+type_register(WidgetSelection)
